/*
 * Copyright 2002 Barnaba Marcello <vjt@azzurra.org>
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2a. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in
 *     the documentation and/or other materials provided with the
 *     distribution.
 * 2b. Redistribution in binary form requires specific prior written
 *     authorization of the maintainer.
 * 
 * 3. All advertising materials mentioning features or use of this
 *    software must display the following acknowledgement:
 *    This product includes software developed by Barnaba Marcello.
 * 
 * 4. The names of the maintainer, developers and contributors may not be
 *    used to endorse or promote products derived from this software
 *    without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE MAINTAINER, DEVELOPERS AND
 * CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL
 * THE DEVELOPERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 * ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

/* $Id: chanserv.c,v 1.56 2003/03/29 01:58:55 vjt Exp $ */

#include "sux.h"
#include "agents.h"
#include "chanserv.h"
#include "sql.h"
#include "main.h"
#include "match.h"
#include "help.h"
#include "parse.h"
#include "network.h"
#include "sux-printf.h"
#include "conf.h"
#include "modules.h"
#include "threads.h"

User *cs;
static MYSQL_RES *cs_res;
static MYSQL_ROW cs_row;

const gchar sux_module_name[] = "ChanServ";
const gchar sux_module_version[] = "0.1.3";
const gchar sux_module_info[] = "Channel Management Services";

/* access functions
 */
static void cs_dispatch_access_action(Channel *, ChanMember *, guint);
static guint cs_get_access_from_row(MYSQL_ROW);
static G_INLINE_FUNC guint cs_get_access_internal(GSList *, guint, guint);
static G_INLINE_FUNC guint cs_get_mask_access_internal(User *, guint);
static G_INLINE_FUNC gboolean cs_check_access(Channel *, ChanMember *);

/* MODE functions
 */
static G_INLINE_FUNC void cs_do_mode(Channel *, gchar *, ChanMember *);
static gboolean cs_mode(User *, gint, gchar **, gchar *);
static gboolean cs_send_modelock(Channel *);

/* channel identification functions
 */
static G_INLINE_FUNC SLink *cs_create_id_chan(gchar *);
static G_INLINE_FUNC SLink *cs_find_id_chan(gchar *, User *);

/* synch functions
 */
static void synch_channel(gchar *, Channel *);
static void cs_initial_synch(void);

/* expiration
 */
static gboolean cs_expire_channels(void);

/* misc functions
 */
static gchar *make_nick_user_host(User *);
static G_INLINE_FUNC gchar *cs_get_akick_reason(Channel *, User *);
static G_INLINE_FUNC gboolean cs_get_chan_and_founder_id(const gchar *, guint *, guint *);
static gboolean cs_update_topic(Channel *);

/* extern`s
 */
REMOTE_TABLE_INSTANCE(user);
REMOTE_TABLE_INSTANCE(channel);

REMOTE_MEMPOOL_INSTANCE(links);

/* for chanserv entry point, look into
 * chanserv-cmd.gperf
 */

gchar *g_module_check_init(GModule *mod)
{
    extern GSList *agents;
    Agent *agent;
    
    agent = g_new0(Agent, 1);
    agent->name = "ChanServ";
    agent->id = SYMBOL_SECTION_CHANSERV;
    agent->u = &cs;

    agents = g_slist_prepend(agents, (gpointer) agent);

    agent_setcallback("CHANSERV", chanserv);

    message_add_callback(MESSAGE_NEWCHAN, SUX_MSG_PRIORITY_NORMAL, (message_func) cs_send_modelock);
    message_add_callback(MESSAGE_JOIN, SUX_MSG_PRIORITY_NORMAL, (message_func) cs_check_access);
    message_add_callback(MESSAGE_TOPIC, SUX_MSG_PRIORITY_NORMAL, (message_func) cs_update_topic);
    message_add_callback(MESSAGE_EXPIRE, SUX_MSG_PRIORITY_NORMAL, (message_func) cs_expire_channels);

    message_add_callback(MESSAGE_CSREQ, SUX_MSG_PRIORITY_NORMAL, (message_func) chanserv);
    
    if(me.boot)
    {
	agent_boot(agent);
	cs_initial_synch();
    }

    g_message_syslog("MDL: Loaded ChanServ module");

    return NULL;
}

void g_module_unload(GModule *mod)
{
    extern GSList *agents;
    GSList *agent;

    agent = g_slist_find_custom(agents, GINT_TO_POINTER(SYMBOL_SECTION_CHANSERV),
	    (GCompareFunc) agent_find);
    agents = g_slist_remove_link(agents, agent);
    agent_shutdown((Agent *)agent->data);

    g_free(agent->data);
    g_slist_free_1(agent);

    message_del_callback(MESSAGE_NEWCHAN, (message_func) cs_send_modelock);
    message_del_callback(MESSAGE_JOIN, (message_func) cs_check_access);
    message_del_callback(MESSAGE_TOPIC, (message_func) cs_update_topic);
    message_del_callback(MESSAGE_EXPIRE, (message_func) cs_expire_channels);

    message_del_callback(MESSAGE_CSREQ, (message_func) chanserv);

    g_message_syslog("MDL: Unloaded ChanServ module");
}

static void cs_initial_synch(void)
{
    time_t synch_time;

    sql_lock_tables();

    synch_time = time(NULL);
    send_out(":%s GLOBOPS :starting channel synch with database", me.name);
    g_hash_table_foreach(_TBL(channel).__hash_tbl, (GHFunc) synch_channel, NULL);
    send_out(":%s GLOBOPS :channel synch completed in %ld second(s)", me.name, time(NULL) - synch_time);

    sql_unlock_tables();
}

static void synch_channel(gchar *dummy, Channel *c)
{
    register GSList *johnny /* walker */;
    register ChanMember *cm;

    cs_send_modelock(c);

    johnny = c->members;
    while(johnny)
    {
	cm = (ChanMember *) johnny->data;
	johnny = g_slist_next(johnny);

	if(cm->u->flags & (FLAGS_MYAGENT|FLAGS_MYCLIENT|FLAGS_SERVER))
	    continue;

	cs_check_access(c, cm);
    }
}

/* help wrapper around sux_help()
 */
void cs_help(User *caller, gint parc, gchar **parv)
{
    sux_help(caller, cs, CHANSERV_HELP_ID, parv ? parv[0] : NULL);
}

/*
 * parv[0] -> channel
 * parv[1] -> password
 * parv[2] -> description
 */
void cs_register(User *caller, gint parc, gchar **parv)
{
    gchar *channel;
    Channel *c;
    ChanMember *cm;

    guint founder_id;

    if(parc < 2)
    {
	send_out(":%s NOTICE %s :Not enough parameters",
		cs->name, caller->name);

	return;
    }

    channel = parv[0];

    sql_query_assert(&cs_res,
	    "SELECT `id` FROM `chans` WHERE `name` = '%s'",
	    channel);

    if(mysql_num_rows(cs_res))
    {
	send_out(":%s NOTICE %s :Channel %s is already registered",
		cs->name, caller->name, channel);

	mysql_free_result(cs_res);
	return;
    }

    mysql_free_result(cs_res);

    c = _TBL(channel).get(channel);
    if(!c)
    {
	send_out(":%s NOTICE %s :Channel %s does not exist !",
		cs->name, caller->name, channel);
	return;
    }

    cm = find_chanmember(c, caller);
    if(!cm)
    {
	send_out(":%s NOTICE %s :You must be into channel %s before trying to register it",
		cs->name, caller->name, channel);
	return;
    }

    if(!(cm->flags & CHFL_CHANOP))
    {
	send_out(":%s NOTICE %s :You must have channel operator status into channel %s in order to register it",
		cs->name, caller->name, channel);
	return;
    }

    sql_query_assert(&cs_res,
	    "SELECT `id` FROM `nicks` WHERE `name` = '%s'",
	    caller->name);

    if(!mysql_num_rows(cs_res))
    {
	send_out(":%s NOTICE %s :Your nick is not registered !",
		cs->name, caller->name);

	mysql_free_result(cs_res);
	return;
    }

    cs_row = mysql_fetch_row(cs_res);
    founder_id = atoi(*cs_row);
    mysql_free_result(cs_res);

    /* ok, proceed to registration
     */

    sql_query_assert(NULL,
	    "INSERT INTO `chans` "
	    "VALUES(0, '%s', %d, '%s', '%s', NOW(), NOW(), '+nrt', '%s', '%s', FROM_UNIXTIME(%ld))",
	    channel, founder_id, parv[1], parv[2], c->topic, c->topic_nick, c->topic_time);

    send_out(":%s NOTICE %s :Channel %s is now registered to you. Your password is %s",
	    cs->name, caller->name, channel, parv[1]);
    
    send_out(":%s MODE %s +nrt", cs->name, c->chname);
    c->mode.mode |= MODE_REGISTERED|MODE_NOPRIVMSGS|MODE_TOPICLIMIT;

    caller->id_chans = g_slist_prepend(caller->id_chans, cs_create_id_chan(c->chname));

    return;

}

/*
 * parv[0] -> channel name
 * parv[1] -> password
 */
void cs_identify(User *caller, gint parc, gchar **parv)
{
    gchar *channel = parv[0];
    gchar *passwd = parv[1];
    
    if(parc < 2)
    {
	send_out(":%s NOTICE %s :Not enough parameters",
			cs->name, caller->name);
	return;
    }

    if(cs_find_id_chan(channel, caller))
    {
	send_out(":%s NOTICE %s :You are already identified to channel %s",
		cs->name, caller->name, channel);
	return;
    }

    sql_query_assert(&cs_res,
	    "SELECT `pass` FROM `chans` WHERE `name` = '%s'",
	    channel);

    if(!mysql_num_rows(cs_res))
    {
	send_out(":%s NOTICE %s :Channel %s is not registered",
		cs->name, caller->name, channel);

	mysql_free_result(cs_res);
	return;
    }

    cs_row = mysql_fetch_row(cs_res);
    if(strcmp(*cs_row, passwd))
    {
	send_out(":%s NOTICE %s :Password supplied for %s is incorrect",
		cs->name, caller->name, channel);
	mysql_free_result(cs_res);
	return;
    }

    mysql_free_result(cs_res);

    send_out(":%s NOTICE %s :Password accepted for %s, you are now identified as the channel founder",
	    cs->name, caller->name, channel);

    caller->id_chans = g_slist_prepend(caller->id_chans, cs_create_id_chan(channel));

    return;
}

/*
 * parv[0] -> channel to query
 */
void cs_info(User *caller, gint parc, gchar **parv)
{
    if(parc < 1)
    {
	send_out(":%s NOTICE %s :Not enough parameters",
		cs->name, caller->name);
	return;
    }

    sql_query_assert(&cs_res,
	    "SELECT `chans`.`name`, `chans`.`desc`, DATE_FORMAT(`chans`.`time_reg`, '%%b %%e %%Y - %%T'), "
	    "DATE_FORMAT(`chans`.`time_used`, '%%b %%e %%Y - %%T'), `chans`.`mlock`, `chans`.`last_topic`, "
	    "`chans`.`last_topic_nick`, DATE_FORMAT(`chans`.`last_topic_time`, '%%b %%e %%Y - %%T'), "
	    "`nicks`.`name`, DATE_FORMAT(NOW(), '%%b %%e %%Y - %%T') "
	    "FROM `chans`, `nicks` "
	    "WHERE `chans`.`founder_id` = `nicks`.`id` AND `chans`.`name` = '%s'",
	    parv[0]);

    if(!mysql_num_rows(cs_res))
    {
	send_out(":%s NOTICE %s :Channel %s is not registered !",
		cs->name, caller->name, parv[0]);
	mysql_free_result(cs_res);

	return;
    }

    cs_row = mysql_fetch_row(cs_res);

    send_out(":%s NOTICE %s :*** Info about %s", cs->name, caller->name, cs_row[0]);
    send_out(":%s NOTICE %s :Founder: %s", cs->name, caller->name, cs_row[8]);
    send_out(":%s NOTICE %s :Description: %s", cs->name, caller->name, cs_row[1]);
    send_out(":%s NOTICE %s :Registration date: %s", cs->name, caller->name, cs_row[2]);
    send_out(":%s NOTICE %s :Last join: %s", cs->name, caller->name, cs_row[3]);
    send_out(":%s NOTICE %s :Time now: %s", cs->name, caller->name, cs_row[9]);
    send_out(":%s NOTICE %s :Last topic: %s", cs->name, caller->name, cs_row[5]);
    send_out(":%s NOTICE %s :Topic changed by: %s on: %s", cs->name, caller->name, cs_row[6], cs_row[7]);
    send_out(":%s NOTICE %s :Modelock: %s", cs->name, caller->name, cs_row[4]);
    send_out(":%s NOTICE %s :*** End of info", cs->name, caller->name);

    mysql_free_result(cs_res);
    
    return;
}

void cs_op(User *caller, gint parc, gchar **parv)
{
    if(cs_mode(caller, parc, parv, "+o"))
    {
	send_out(":%s NOTICE @%s :%s used OP on %s",
		cs->name, parv[0], caller->name, parv[1]);
    }
}

void cs_deop(User *caller, gint parc, gchar **parv)
{
    if(cs_mode(caller, parc, parv, "-o"))
    {
	send_out(":%s NOTICE @%s :%s used DEOP on %s",
		cs->name, parv[0], caller->name, parv[1]);
    }
}

void cs_voice(User *caller, gint parc, gchar **parv)
{
    if(cs_mode(caller, parc, parv, "+v"))
    {
	send_out(":%s NOTICE @%s :%s used VOICE on %s",
		cs->name, parv[0], caller->name, parv[1]);
    }
}

void cs_devoice(User *caller, gint parc, gchar **parv)
{
    if(cs_mode(caller, parc, parv, "-v"))
    {
	send_out(":%s NOTICE @%s :%s used DEVOICE on %s",
		cs->name, parv[0], caller->name, parv[1]);
    }
}

void cs_invite(User *caller, gint parc, gchar **parv)
{
    Channel *c;

    if(parc < 1)
    {
	send_out(":%s NOTICE %s :Not enough parameters",
		cs->name, caller->name);
	return;
    }

    c = _TBL(channel).get(parv[0]);
    if(!c)
    {
	send_out(":%s NOTICE %s :Channel %s does not exist !",
		cs->name, caller->name, parv[0]);
	return;
    }

    if(find_chanmember(c, caller))
    {
	send_out(":%s NOTICE %s :You are already on channel %s",
		cs->name, caller->name, c->chname);
	return;
    }

    switch(cs_get_access(c->chname, caller))
    {
	case CS_ERR_NOTREG:
	    send_out(":%s NOTICE %s :Channel %s is not registered !",
		    cs->name, caller->name, c->chname);
	    break;
	case CS_ACCESS_VOP ... CS_ACCESS_FOUNDER:
	    send_out(":%s INVITE %s %s",
		    cs->name, caller->name, c->chname);
	    break;
	case CS_ERR_NOACCESS:
	default:
	    send_out(":%s NOTICE %s :Access denied",
		    cs->name, caller->name);
	    break;
    }

    return;
}

/* parv[0] -> channel
 * parv[1] -> nick
 */
static gboolean cs_mode(User *caller, gint parc, gchar **parv, gchar *mode)
{
    Channel *c;
    User *u;
    ChanMember *cm;

    if(parc < 2)
    {
	send_out(":%s NOTICE %s :Not enough parameters",
		cs->name, caller->name);
	return FALSE;
    }

    c = _TBL(channel).get(parv[0]);
    if(!c)
    {
	send_out(":%s NOTICE %s :Channel %s does not exist !",
		cs->name, caller->name, parv[0]);
	return FALSE;
    }

    u = _TBL(user).get(parv[1]);
    if(!u)
    {
	send_out(":%s NOTICE %s :User %s unknown",
		cs->name, caller->name, parv[1]);
	return FALSE;
    }

    cm = find_chanmember(c, u);
    if(!cm)
    {
	send_out(":%s NOTICE %s :User %s is not on channel %s",
		cs->name, caller->name, u->name, c->chname);
	return FALSE;
    }

    switch(*(mode + 1))
    {
	case 'o':
	    if((*mode == '+' && cm->flags & CHFL_CHANOP) ||
		    (*mode == '-' && !(cm->flags & CHFL_CHANOP)))
		return FALSE;
	    break;

	case 'v':
	    if((*mode == '+' && cm->flags & CHFL_VOICE) ||
		    (*mode == '-' && !(cm->flags & CHFL_VOICE)))
		return FALSE;
	    break;

	default:
	    g_critical("Unknown mode passed to cs_mode() (%s)",
		    mode);
    }

    /* access check
     */

    switch(cs_get_access(c->chname, caller))
    {
	case CS_ERR_NOTREG:
	    send_out(":%s NOTICE %s :Channel %s is not registered !",
		    cs->name, caller->name, c->chname);
	    break;
	case CS_ACCESS_AOP ... CS_ACCESS_FOUNDER:
	    cs_do_mode(c, mode, cm);
	    return TRUE;
	case CS_ERR_NOACCESS:
	default:
	    send_out(":%s NOTICE %s :Access denied", cs->name, caller->name);
	    break;
    }

    return FALSE;
}


static G_INLINE_FUNC void cs_do_mode(Channel *c, gchar *mode, ChanMember *cm)
{
    send_out(":%s MODE %s %s %s",
	    cs->name, c->chname, mode, cm->u->name);

    switch(*(mode + 1))
    {
	case 'o':
	    if(*mode == '+')
		cm->flags |= CHFL_CHANOP;
	    else
		cm->flags &= ~CHFL_CHANOP;
	    break;

	case 'v':
	    if(*mode == '+')
		cm->flags |= CHFL_VOICE;
	    else
		cm->flags &= ~CHFL_VOICE;
	    break;
    }
}

gint cs_get_access(gchar *chname, User *u)
{
    guint chan_id, founder_id, access;
    GSList *johnny /* walker */;
    
    if(cs_find_id_chan(chname, u))
    {
	return CS_ACCESS_FOUNDER;
    }

    /* this channel is registered ? */
    if(!cs_get_chan_and_founder_id(chname, &chan_id, &founder_id))
    {
	return CS_ERR_NOTREG;
    }

    /* loop the id nicks of this user */
    johnny = u->id_nicks;
    while(johnny)
    {
	if((access = cs_get_access_internal(johnny, chan_id, founder_id)))
	{
	    return access;
	}

	johnny = g_slist_next(johnny);
    }

    /* search for masks .. */
    return cs_get_mask_access_internal(u, chan_id);
}

static G_INLINE_FUNC guint cs_get_access_internal(GSList *id_nick, guint chan_id, guint founder_id)
{
    guint nick_id;
    gchar *nick = ((SLink *)id_nick->data)->v.cp;
    gushort access = 0;

    /* is this the channel founder ? */
    sql_query_assert_ret(0, &cs_res,"SELECT `id` FROM `nicks` WHERE `nicks`.`name` = '%s'", nick);

    cs_row = mysql_fetch_row(cs_res);

    nick_id = atoi(*cs_row);
    mysql_free_result(cs_res);

    if(nick_id == founder_id)
    {
	/* okay, the channel founder has joined ! */	    
	return CS_ACCESS_FOUNDER;
    }

    /* search in access tables */
    sql_query_assert_ret(0, &cs_res,
	    "SELECT `cfounder`.`nick_id`, `sop`.`nick_id`, "
	    "`aop`.`nick_id`, `vop`.`nick_id`, `akick`.`nick_id` "
	    "FROM `nicks` "
	    "LEFT JOIN `cfounder` ON ( `cfounder`.`flags` = 'nick' AND "
	    	"`cfounder`.`chan_id` = '%d' AND `cfounder`.`nick_id` = `nicks`.`id` ) "
	    "LEFT JOIN `sop` ON ( `sop`.`flags` = 'nick' AND "
	    	"`sop`.`chan_id` = '%d' AND `sop`.`nick_id` = `nicks`.`id` ) "
	    "LEFT JOIN `aop` ON ( `aop`.`flags` = 'nick' AND "
	    	"`aop`.`chan_id` = '%d' AND `aop`.`nick_id` = `nicks`.`id` ) "
	    "LEFT JOIN `vop` ON ( `vop`.`flags` = 'nick' AND "
	    	"`vop`.`chan_id` = '%d' AND `vop`.`nick_id` = `nicks`.`id` ) "
	    "LEFT JOIN `akick` ON ( `akick`.`flags` = 'nick' AND "
	    	"`akick`.`chan_id` = '%d' AND `akick`.`nick_id` = `nicks`.`id` ) "
	    "WHERE `nicks`.`id` = '%d'",
	        chan_id, chan_id, chan_id, chan_id, chan_id, nick_id);

    if(!mysql_num_rows(cs_res))
    {
	/* no access */
	mysql_free_result(cs_res);

	return 0;
    }

    cs_row = mysql_fetch_row(cs_res);
    access = cs_get_access_from_row(cs_row);
    mysql_free_result(cs_res);
    
    return access;
}

static G_INLINE_FUNC guint cs_get_mask_access_internal(User *who, guint chan_id)
{
    guint access;

    sql_query_assert_ret(0, &cs_res,
	    "SELECT `cfounder`.`nick_id`, `sop`.`nick_id`, "
	    	"`aop`.`nick_id`, `vop`.`nick_id`, `akick`.`nick_id` "
	    "FROM `masks` "
	    "LEFT JOIN `cfounder` ON ( `cfounder`.`flags` = 'mask' AND "
	    		"`cfounder`.`chan_id` = '%d' AND `cfounder`.`nick_id` = `masks`.`id` ) "
	    "LEFT JOIN `sop` ON ( `sop`.`flags` = 'mask' AND "
	    		"`sop`.`chan_id` = '%d' AND `sop`.`nick_id` = `masks`.`id` ) "
	    "LEFT JOIN `aop` ON ( `aop`.`flags` = 'mask' AND "
			"`aop`.`chan_id` = '%d' AND `aop`.`nick_id` = `masks`.`id` ) "
	    "LEFT JOIN `vop` ON ( `vop`.`flags` = 'mask' AND "
	    		"`vop`.`chan_id` = '%d' AND `vop`.`nick_id` = `masks`.`id` ) "
	    "LEFT JOIN `akick` ON ( `akick`.`flags` = 'mask' AND "
	    		"`akick`.`chan_id` = '%d' AND `akick`.`nick_id` = `masks`.`id` ) "
	    "WHERE '%s' LIKE `masks`.`mask` AND ( "
	    	"`cfounder`.`nick_id` IS NOT NULL OR `sop`.`nick_id` IS NOT NULL OR "
			"`aop`.`nick_id` IS NOT NULL OR `vop`.`nick_id` IS NOT NULL OR "
			"`akick`.`nick_id` IS NOT NULL )",
	    chan_id, chan_id, chan_id, chan_id, chan_id, make_nick_user_host(who));

    if(!mysql_num_rows(cs_res))
    {
	/* no access .. deop him.
	 */
	mysql_free_result(cs_res);
	return CS_ERR_NOACCESS;
    }

    cs_row = mysql_fetch_row(cs_res);
    access = cs_get_access_from_row(cs_row);
    mysql_free_result(cs_res);

    return access;
}

static G_INLINE_FUNC gboolean cs_get_chan_and_founder_id(const gchar *chname, guint *chan_id, guint *founder_id)
{
    sql_query_assert_ret(FALSE, &cs_res, 
	    "SELECT `id`, `founder_id` FROM `chans` WHERE chans.name = '%s'",
	    chname);

    if(!mysql_num_rows(cs_res))
    {
	/* no .. */
	mysql_free_result(cs_res);
	return FALSE;
    }

    /* yes .. save the chan id for faster processing */
    cs_row = mysql_fetch_row(cs_res);
    *chan_id = atoi(cs_row[0]);
    *founder_id = atoi(cs_row[1]);

    mysql_free_result(cs_res);

    return TRUE;
}

G_INLINE_FUNC gboolean cs_check_access(Channel *c, ChanMember *cm)
{
    User *u;
    GSList *johnny /* walker */;
    guint access, curr_flags;

    u = cm->u;
    curr_flags = cm->flags;
    johnny = u->id_nicks;

    if((access = cs_get_access(c->chname, u)) > CS_ERR_NOACCESS)
    {
	cs_dispatch_access_action(c, cm, access);
	/* update last join
	 */
	sql_query_assert_ret(FALSE, NULL,
		"UPDATE `chans` SET `time_reg` = `time_reg`, `time_used` = NOW() "
		"WHERE `name` = '%s'",
		c->chname);
    }
    
    /* done ! */
    return TRUE;
}

static void cs_dispatch_access_action(Channel *c, ChanMember *cm, guint access)
{
    User *u = cm->u;

    switch(access)
    {
	case CS_ACCESS_AOP ... CS_ACCESS_FOUNDER:
	    if(!(cm->flags & CHFL_CHANOP))
	    {
		send_out(":%s MODE %s +o %s",
			cs->name, c->chname, u->name);
		cm->flags |= CHFL_CHANOP;
	    }
	    return;

	case CS_ACCESS_VOP:
	    if(!(cm->flags & CHFL_VOICE))
	    {
		send_out(":%s MODE %s +v %s",
			cs->name, c->chname, u->name);
		cm->flags |= CHFL_VOICE;
	    }
	    return;

	case CS_ACCESS_AKICK:
	    send_out(":%s MODE %s +b *!*@%s",
		    cs->name, c->chname, u->host);
	    send_out(":%s KICK %s %s :%s",
		    cs->name, c->chname, u->name, cs_get_akick_reason(c, u));
	    remove_user_from_channel(c, u);
	    return;
    }
}

static G_INLINE_FUNC SLink *cs_find_id_chan(gchar *chname, User *u)
{
    GSList *johnny = u->id_chans; /* walker */

    while(johnny)
    {
	if(!mycmp(((SLink *)johnny->data)->v.cp, chname))
	{
	    return ((SLink *) johnny->data);
	}

	johnny = g_slist_next(johnny);
    }

    return NULL;
}

static G_INLINE_FUNC SLink *cs_create_id_chan(gchar *chname)
{
    SLink *ret;

    ret = g_mem_chunk_alloc0(_MPL(links));
    ret->v.cp = g_strdup(chname);
    ret->flags = 0;

    return ret;
}

static gboolean cs_update_topic(Channel *c)
{
    if(!(c->mode.mode & MODE_REGISTERED))
    {
	return TRUE;
    }

    sql_query_assert_ret(FALSE, NULL,
	    "UPDATE `chans` SET "
	    "`last_topic` = '%s', `last_topic_nick` = '%s', `last_topic_time` = FROM_UNIXTIME('%ld') "
	    "WHERE `name` = '%s'",
	    c->topic, c->topic_nick, c->topic_time, c->chname);

    return TRUE;
}

static gboolean cs_send_modelock(Channel *c)
{
    gchar *mode_arr[3];
    gchar mode_str[32], *mode_args[2];
    gchar *m = mode_str, *s, *p;
    guint count, change = 0, l, i = 0, arg = 1, newarg = 0;

    static const gint flags[] = 
    {
	MODE_NOPRIVMSGS, 'n', MODE_REGISTERED, 'r', MODE_TOPICLIMIT, 't',
	MODE_PRIVATE, 'p', MODE_SECRET, 's', MODE_MODERATED, 'm', 
	MODE_REGONLY, 'R', MODE_INVITEONLY, 'i', MODE_NOCOLOR, 'c',
	MODE_OPERONLY, 'O', MODE_MODREG, 'M', MODE_NONICKCHG, 'd',
	0x0, 0x0
    };
    
    /* this channel is registered ? */
    sql_query_assert_ret(FALSE, &cs_res,
	    "SELECT `mlock` FROM `chans` WHERE `chans`.`name` = '%s'",
	    c->chname);

    if(!mysql_num_rows(cs_res))
    {
	/* no .. */
	mysql_free_result(cs_res);
	return TRUE;
    }

    /* yes ..  */
    cs_row = mysql_fetch_row(cs_res);
    
    if(cs_row[0] == NULL)
    {
	/* no modelock
	 */
	mysql_free_result(cs_res);
	return TRUE;
    }

    my_g_strsplit(cs_row[0], ' ', 3, &count, mode_arr);
    *mode_str = '\0';
    mode_args[0] = mode_args[1] = NULL;

    for(s = mode_arr[0]; *s; s++)
    {
	switch(*s)
	{
	    case '+':
		change = *m++ = '+';
		break;
	    case '-':
		change = *m++ = '-';
		break;
	    case 'l':
		switch(change)
		{
		    case '+':
			p = mode_arr[arg++];
			l = atoi(p);
			if(c->mode.mode & MODE_LIMIT)
			{
			    if(l != c->mode.limit)
			    {
				c->mode.limit = l;
				
				*m++ = 'l';
				mode_args[newarg++] = p;
			    }
			}
			else
			{
			    c->mode.mode |= MODE_LIMIT;
			    c->mode.limit = l;

			    *m++ = 'l';
			    mode_args[newarg++] = p;
			}
			break;
		    case '-':
			if(c->mode.mode & MODE_LIMIT)
			{
			    c->mode.mode &= ~MODE_LIMIT;
			    c->mode.limit = 0;

			    *m++ = 'l';
			}
			break;
		}
		break;

	    case 'k':
		switch(change)
		{
		    case '+':
			p = mode_arr[arg++];
			if(c->mode.mode & MODE_KEY)
			{
			    if(mycmp(p, c->mode.key))
			    {
				g_strlcpy(c->mode.key, p, KEYLEN);

				*m++ = 'k';
				mode_args[newarg++] = p;
			    }
			}
			else
			{
			    c->mode.mode |= MODE_KEY;
			    g_strlcpy(c->mode.key, p, KEYLEN);

			    *m++ = 'k';
			    mode_args[newarg++] = p;
			}
			break;
		    case '-':
			if(c->mode.mode & MODE_KEY)
			{
			    c->mode.mode &= ~MODE_KEY;
			    *c->mode.key = '\0';

			    *m++ = 'k';
			}
			break;
		}
		break;

	    default:
		for(i = 1; flags[i] != 0; i += 2)
		{
		    if(*s == (gchar)flags[i])
		    {
			if(change == '+' && !(c->mode.mode & flags[i - 1]))
			{
			    c->mode.mode |= flags[i - 1];
			    *m++ = (gchar) flags[i];
			}
			else if(change == '-' && (c->mode.mode & flags[i - 1]))
			{
			    c->mode.mode &= ~flags[i - 1];
			    *m++ = (gchar) flags[i];
			}
			
			break;
		    }
		}
	}
    }

    *m = '\0';
    if(*mode_str)
    {
	send_out(":%s MODE %s %s %s %s", cs->name, c->chname, mode_str,
		mode_args[0] ? mode_args[0] : "",
		mode_args[1] ? mode_args[1] : "");
    }

    mysql_free_result(cs_res);

    return TRUE;
}

static guint cs_get_access_from_row(MYSQL_ROW row)
{
    guint access = 0x0;

    if(row[0])
    {
	access = CS_ACCESS_COFOUNDER;
    }
    else if(row[1])
    {
	access = CS_ACCESS_SOP;
    }
    else if(row[2])
    {
	access = CS_ACCESS_AOP;
    }
    else if(row[3])
    {
	access = CS_ACCESS_VOP;
    }
    else if(row[4])
    {
	access = CS_ACCESS_AKICK;
    }

    return access;
}

gchar *make_nick_user_host(User *u)
{
    /* nick!user@host\0
     */
    static gchar retbuf[NICKLEN + 1 + USERLEN + 1 + HOSTLEN + 1];

    sux_snprintf(retbuf, sizeof(retbuf), "%s!%s@%s", u->name, u->username, u->host);

    return retbuf;
}

static G_INLINE_FUNC gchar *cs_get_akick_reason(Channel *c, User *u)
{
    static gchar ret[256];

    sql_query_assert_ret(NULL, &cs_res, "SELECT `reason` FROM `akick` "
	    "LEFT JOIN `chans` ON ( `akick`.`chan_id` = `chans`.`id` ) "
	    "LEFT JOIN `nicks` ON ( `akick`.`nick_id` = `nicks`.`id` ) "
	    "WHERE `akick`.`flags` = 'nick' AND `chans`.`name` = '%s' AND `nicks`.`name` = '%s'",
	    c->chname, u->name);

    if(!mysql_num_rows(cs_res))
    {
	/* try mask access
	 */
	mysql_free_result(cs_res);
	sql_query_assert_ret(NULL, &cs_res, "SELECT `reason` FROM `akick` "
		"LEFT JOIN `chans` ON ( `akick`.`chan_id` = `chans`.`id` ) "
		"LEFT JOIN `masks` ON ( `akick`.`nick_id` = `masks`.`id` ) "
		"WHERE `akick`.`flags` = 'mask' AND `chans`.`name` = '%s' AND '%s' LIKE `masks`.`mask`",
		c->chname, make_nick_user_host(u));

	if(!mysql_num_rows(cs_res))
	{
	    /* this should not happen
	     */
	    g_critical("AKICK access not found, but present in user list !!");
	    return NULL;
	}
    }

    cs_row = mysql_fetch_row(cs_res);
	    
    strcpy(ret, cs_row[0]);
	    
    mysql_free_result(cs_res);

    return ret;
}

/* channel expiration
 */
static gboolean cs_expire_channels(void)
{
    static guint expire_no = 0;
    time_t expire_started = time(NULL);
    guint deleted_chans;

    send_out(":%s GLOBOPS :Starting channel expire #%d",
	    cs->name, ++expire_no);
    
    sql_query_assert_ret(FALSE, &cs_res, "SELECT `id`, `name` FROM `chans` "
	    "WHERE TO_DAYS(NOW()) - TO_DAYS(`time_used`) > 40");

    while((cs_row = mysql_fetch_row(cs_res)))
    {
	gchar *chan_id = cs_row[0];
	const gchar *channel_tables[] = 
	{
	    "akick",
	    "aop",
	    "cfounder",
	    "sop",
	    "vop",
	    NULL
	};
	gint i;

	send_out(":%s PRIVMSG #sux :Expiring channel %s",
		cs->name, cs_row[1]);

	for(i = 0; channel_tables[i]; i++)
	{
	    sql_query_assert_ret(FALSE, NULL, "DELETE FROM `%s` WHERE `chan_id` = %s",
		    channel_tables[i], chan_id);
	}
    }

    mysql_free_result(cs_res);

    sql_query_assert_ret(FALSE, NULL, "DELETE FROM `chans` "
	    "WHERE TO_DAYS(NOW()) - TO_DAYS(`time_used`) > 40");

    deleted_chans = sql_affected_rows();

    send_out(":%s GLOBOPS :Channel expire completed in %ld secs. Deleted %d channels",
	    cs->name, time(NULL) - expire_started, deleted_chans);

    return TRUE;
}

/* EOF
 */
