/*
 * Copyright 2002 Barnaba Marcello <vjt@azzurra.org>
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2a. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in
 *     the documentation and/or other materials provided with the
 *     distribution.
 * 2b. Redistribution in binary form requires specific prior written
 *     authorization of the maintainer.
 * 
 * 3. All advertising materials mentioning features or use of this
 *    software must display the following acknowledgement:
 *    This product includes software developed by Barnaba Marcello.
 * 
 * 4. The names of the maintainer, developers and contributors may not be
 *    used to endorse or promote products derived from this software
 *    without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE MAINTAINER, DEVELOPERS AND
 * CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL
 * THE DEVELOPERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 * ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

/* $Id: conf.c,v 1.12 2003/03/23 18:10:10 vjt Exp $ */

#include "sux.h"
#include "log.h"
#include "conf.h"
#include "main.h"

static GNode *conf;

/* section {
 * 	item = "value";
 * 	item2 = "value2";
 * 	...
 * }
 *
 * section2 {
 * 	item = "value";
 * 	item2 = "value2";
 * 	....
 * }
 */
static const struct section
{
    gchar *name;
    guint tok;
} sections[] = {
    { "{",		SYMBOL_SECTION_START },
    { "}",		SYMBOL_SECTION_END },
    { "globals",	SYMBOL_SECTION_MAIN },
    { "sql",		SYMBOL_SECTION_SQL },
    { "chanserv",	SYMBOL_SECTION_CHANSERV },
    { "nickserv",	SYMBOL_SECTION_NICKSERV },
    { "operserv",	SYMBOL_SECTION_OPERSERV },
    { "rootserv",	SYMBOL_SECTION_ROOTSERV },
    { "memoserv",	SYMBOL_SECTION_MEMOSERV },
    { "noticer",	SYMBOL_SECTION_NOTICER },
    { "modules",	SYMBOL_SECTION_MODULES },
    { NULL,		0 }

}, main_section_symbols[] = {
    { "my_name",	SYMBOL_MY_NAME },
    { "my_pass",	SYMBOL_MY_PASS },
    { "my_info",	SYMBOL_MY_INFO },
    { "his_name",	SYMBOL_UPLINK_NAME },
    { "his_pass",	SYMBOL_UPLINK_PASS },
    { "his_host",	SYMBOL_UPLINK_HOST },
    { "his_port",	SYMBOL_UPLINK_PORT },
    { NULL,		0 }

}, agent_section_symbols[] = {
    { "name",		SYMBOL_AGENT_NAME },
    { "username",	SYMBOL_AGENT_USERNAME },
    { "host",		SYMBOL_AGENT_HOST },
    { "gcos",		SYMBOL_AGENT_INFO },
    { NULL,		0 }

}, sql_section_symbols[] = {
    { "host",		SYMBOL_SQL_HOST },
    { "user",		SYMBOL_SQL_USER },
    { "pass",		SYMBOL_SQL_PASS },
    { "db",		SYMBOL_SQL_DB },
    { "port",		SYMBOL_SQL_PORT },
    { NULL,		0 }
}, modules_section_symbols[] = {
    { "load",		SYMBOL_MODULE_LOAD },
    { NULL,		0 }
};

static const struct section_group
{
    const struct section *name;
    guint scope;
} section_groups[] = {
   { sections,			SCOPE_FILE },
   { main_section_symbols,	SCOPE_MAIN_SECTION },
   { agent_section_symbols,	SCOPE_AGENT_SECTION },
   { sql_section_symbols,	SCOPE_SQL_SECTION },
   { modules_section_symbols,	SCOPE_MODULES_SECTION },
   { NULL,			0 }
};

static GScannerConfig *configure_scanner(void)
{
    GScannerConfig *cfg = g_new0(GScannerConfig, 1);
    
    cfg->cset_skip_characters = " \t\r\n";
    cfg->cset_identifier_first = G_CSET_a_2_z;
    cfg->cset_identifier_nth = G_CSET_a_2_z "_0123456789" G_CSET_A_2_Z;
    cfg->cpair_comment_single = "#\n";
    
    cfg->case_sensitive = FALSE;
    
    cfg->skip_comment_multi = TRUE;
    cfg->skip_comment_single = TRUE;
    cfg->scan_comment_multi = TRUE;
    
    cfg->scan_identifier = TRUE;
    cfg->scan_identifier_1char = FALSE;
    cfg->scan_identifier_NULL = TRUE;

    cfg->scan_symbols = TRUE;
    cfg->scan_binary = FALSE;
    cfg->scan_octal = FALSE;
    cfg->scan_float = FALSE;
    cfg->scan_hex = TRUE;
    cfg->scan_hex_dollar = FALSE;
    cfg->scan_string_sq = FALSE;
    cfg->scan_string_dq = TRUE;

    cfg->numbers_2_int = TRUE;
    cfg->int_2_float = FALSE;
    cfg->identifier_2_string = FALSE;
    cfg->char_2_token = TRUE;
    cfg->symbol_2_token = TRUE;
    cfg->scope_0_fallback = FALSE;

    return cfg;
}

static void initialize_symbols(GScanner *lex)
{
    const struct section_group *sec;
    const struct section *sym;

    sec = section_groups;
    while(sec->name)
    {
	sym = sec->name;
	while(sym->name)
	{
	    g_scanner_scope_add_symbol(lex, sec->scope,
		    sym->name, GINT_TO_POINTER(sym->tok));
	    sym++;
	}
	sec++;
    }
}

static GScanner *initialize_scanner(gchar *file)
{
    if(access(file, R_OK | F_OK) == 0)
    {
	GScannerConfig *cfg;
	GScanner *lex;
	gint fd;
	
	fd = open(file, O_RDONLY);

	if(fd < 0)
	{
	    g_errno_critical("Cannot open config file");
	    return NULL;
	}
	
	cfg = configure_scanner();
	lex = g_scanner_new(cfg);

	if(!lex)
	    return NULL;

	g_scanner_input_file(lex, fd);
	lex->input_name = file;

	initialize_symbols(lex);

	g_scanner_set_scope(lex, SCOPE_FILE);
	
	return lex;
    }
    g_errno_critical("Cannot access configuration file %s", file);

    return NULL;
}

static guint parse_row(GScanner *lex)
{
    GTokenType symbol;
    GTokenValue value;
    static guint scope = SCOPE_FILE;

    static GNode *root;

    ConfigItem *conf_item;
    
    symbol = g_scanner_get_next_token(lex);
    value = g_scanner_cur_value(lex);

    switch(scope)
    {
	case SCOPE_FILE:
	    if(symbol == G_TOKEN_EOF)
		return G_TOKEN_NONE;

	    /* expect sectionname { */
	    if(symbol < FIRST_SECTION_SYMBOL || 
		    symbol > LAST_SECTION_SYMBOL)
	    {
		return G_TOKEN_SYMBOL;
	    }

	    /* expect '{' */
	    g_scanner_get_next_token(lex);
	    if(lex->token != '{')
	    {
		return '{';
	    }

	    conf_item = g_new0(ConfigItem, 1);
	    conf_item->id = value.v_int;
	    conf_item->value = NULL;
	    
	    root = g_node_insert(conf, -1, g_node_new(conf_item));

	    switch(symbol)
	    {
		case SYMBOL_SECTION_MAIN:
		    g_scanner_set_scope(lex, SCOPE_MAIN_SECTION);
		    scope = SCOPE_MAIN_SECTION;

		    return G_TOKEN_NONE;

		case SYMBOL_SECTION_CHANSERV:
		case SYMBOL_SECTION_NICKSERV:
		case SYMBOL_SECTION_MEMOSERV:
		case SYMBOL_SECTION_OPERSERV:
		case SYMBOL_SECTION_ROOTSERV:
		case SYMBOL_SECTION_NOTICER:
		    g_scanner_set_scope(lex, SCOPE_AGENT_SECTION);
		    scope = SCOPE_AGENT_SECTION;

		    return G_TOKEN_NONE;

		case SYMBOL_SECTION_SQL:
		    g_scanner_set_scope(lex, SCOPE_SQL_SECTION);
		    scope = SCOPE_SQL_SECTION;

		    return G_TOKEN_NONE;

		case SYMBOL_SECTION_MODULES:
		    g_scanner_set_scope(lex, SCOPE_MODULES_SECTION);
		    scope = SCOPE_MODULES_SECTION;
		    
		    return G_TOKEN_NONE;

		default:
		    return G_TOKEN_SYMBOL;
		    break;
	    }

	case SCOPE_MAIN_SECTION:
	case SCOPE_AGENT_SECTION:
	case SCOPE_SQL_SECTION:
	case SCOPE_MODULES_SECTION:
	    /* maybe a '}' */
	    
	    if(symbol == '}')
	    {
		g_scanner_set_scope(lex, SCOPE_FILE);
		scope = SCOPE_FILE;
		root = conf;
		return G_TOKEN_NONE;
	    }

	    /* expect '=' */
	    g_scanner_get_next_token(lex);
	    if(lex->token != '=')
	    {
		return '=';
	    }

	    g_scanner_get_next_token(lex);
	    if(lex->token != G_TOKEN_STRING)
	    {
		return G_TOKEN_STRING;
	    }

	    conf_item = g_new0(ConfigItem, 1);
	    conf_item->id = value.v_int;
	    conf_item->value = g_strdup(lex->value.v_string);
	    
	    g_node_insert(root, -1, g_node_new(conf_item));

	    g_scanner_get_next_token(lex);
	    if(lex->token != ';')
	    {
		return ';';
	    }

	    return G_TOKEN_NONE;

	default:
	    g_critical("Unknown scope id %d (huh ?)", scope);
    }

    g_return_val_if_reached(G_TOKEN_ERROR);
}

static G_INLINE_FUNC gboolean find_conf_id_func(GNode *node, gpointer data)
{
    ConfigItem *item = (ConfigItem *)node->data;
    register gpointer *d = data;
    guint id = GPOINTER_TO_INT(*d);

    if(item->id == id)
    {
	*(++d) = item->value;
	return TRUE;
    }

    return FALSE;
}

static G_INLINE_FUNC gboolean find_section_id_func(GNode *node, gpointer data)
{
    ConfigItem *item = (ConfigItem *)node->data;
    register gpointer *d = data;
    guint id = GPOINTER_TO_INT(*d);

    if(item->id == id)
    {
	*(++d) = node;
	return TRUE;
    }

    return FALSE;
}

gchar *find_conf_id(GNode *root, guint id)
{
    gpointer d[2];

    d[0] = GINT_TO_POINTER(id);
    d[1] = NULL;

    if(root == NULL)
	root = conf;

    g_node_traverse(root, G_PRE_ORDER, G_TRAVERSE_LEAFS, -1,
	    find_conf_id_func, d);

    return d[1];
}

GNode *find_section_id(guint id)
{
    gpointer d[2];

    d[0] = GINT_TO_POINTER(id);
    d[1] = NULL;

    g_node_traverse(conf, G_PRE_ORDER, G_TRAVERSE_ALL, -1,
	    find_section_id_func, d);

    return d[1];
}

static gboolean parse_loop(GScanner *lex)
{
    guint expected_token = G_TOKEN_NONE;
    gboolean retval = TRUE;

    while(!g_scanner_eof(lex))
    {
	expected_token = parse_row(lex);
	if(expected_token != G_TOKEN_NONE)
	{
	    g_scanner_unexp_token(lex, expected_token, NULL, NULL, NULL, NULL, TRUE);
	    retval = FALSE;
	    break;
	}
    }
    
    g_scanner_destroy(lex);

    return retval;
}

void setup_configuration()
{
    extern options_t options;
    GScanner *lex;
    
    lex = initialize_scanner(options.conffile);
    if(!lex)
    {
	g_critical("Failed to initialize the lexical scanner");
    }
    
    conf = g_node_new(PACKAGE_STRING);
    if(!parse_loop(lex))
    {
	g_critical("Failed to parse configuration file");
    }
}
