/*
 * Copyright 2002 Barnaba Marcello <vjt@azzurra.org>
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2a. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in
 *     the documentation and/or other materials provided with the
 *     distribution.
 * 2b. Redistribution in binary form requires specific prior written
 *     authorization of the maintainer.
 * 
 * 3. All advertising materials mentioning features or use of this
 *    software must display the following acknowledgement:
 *    This product includes software developed by Barnaba Marcello.
 * 
 * 4. The names of the maintainer, developers and contributors may not be
 *    used to endorse or promote products derived from this software
 *    without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE MAINTAINER, DEVELOPERS AND
 * CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL
 * THE DEVELOPERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 * ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

/* $Id: main.c,v 1.19 2003/03/29 01:58:55 vjt Exp $ */

#define MAIN
#include "sux.h"
#include "main.h"
#include "parse.h"
#include "network.h"
#include "log.h"
#include "h.h"
#include "threads.h"
#include "setproctitle.h"
#include "log.h"
#include "conf.h"
#include "sql.h"
#include "modules.h"

static void help(gchar *);
static void parse_argc_argv(gint, gchar **);
static void lookup_my_info(void);

static gint run(gint, gchar **);
static gint spawn(gint, gchar **);

options_t options;
static struct option long_options[] =
{
    {"conf", 1, 0, 'c'},
    {"mod-dir", 1, 0, 'm'},
    {"nodetach", 0, 0, 'd'},
    {"help", 0, 0, 'h'},
    {NULL, 0, 0, 0}
};

static void help(gchar *myname)
{
    g_printf("Usage: %s [OPTION] ...\n"
	    "Sux Services " PACKAGE_VERSION ", a multithreaded and modular\n"
	    "implementation of standard IRC services.\n"
	    "\n"
	    "  -c, --conf=FILE        Specify config file path.\n"
	    "  -m, --mod-dir=PATH     Specify modules directory.\n"
	    "  -d, --nodetach         Do not detach from controlling terminal.\n"
	    "  -h, --help             This crap.\n"
	    "\n"
	    "Report bugs to <" PACKAGE_BUGREPORT ">\n",
	    myname);

    return;
}

static void parse_argc_argv(gint argc, gchar **argv)
{
    gint option_index = 0;
    gint c;

    options.conffile = SUX_CONFIG_FILE;
    options.moddir = SUX_MODULES_DIR;
    options.nodetach = FALSE;

    while((c = getopt_long(argc, argv, "c:m:dh", long_options, &option_index)) != -1)
    {
	switch(c)
	{
	    case 'c':
		options.conffile = g_strdup(optarg);
		break;
	    case 'm':
		options.moddir = g_strdup(optarg);
		break;
	    case 'd':
		options.nodetach = TRUE;
		break;
	    case 'h':
		help(argv[0]);
		exit(EXIT_SUCCESS);
		break;
	    case '?':
		exit(EXIT_FAILURE);
		break;
	    default:
		g_warning("Unknown option character `%c'", c);
		break;
	}
    }

    return;
}

static void lookup_my_info(void)
{
    gchar *name, *pass, *info;
    name = find_conf_id(NULL, SYMBOL_MY_NAME);
    pass = find_conf_id(NULL, SYMBOL_MY_PASS);
    info = find_conf_id(NULL, SYMBOL_MY_INFO);

    if(!name || !pass || !info)
    {
	g_critical("Incomplete config file");
    }

    g_strlcpy(me.name, name, sizeof(me.name));
    g_strlcpy(me.pass, pass, sizeof(me.pass));
    g_strlcpy(me.info, info, sizeof(me.info));
}

static gint run(gint argc, char **argv)
{
    open_syslog();
    log_set_tty_wrapper();
	
    /* load the configuration file
     */
    setup_configuration();

    lookup_my_info();

    g_message_syslog("%s booting ...", PACKAGE_STRING);

    /* make glib thread-safe
     */
    g_thread_init(NULL);

    if(!sql_connect())
	g_critical_syslog("Cannot connect to the SQL server");
	
    if(!net_connect())
	g_critical_syslog("Cannot connect to the IRC server");

    initproctitle(argc, argv);
    setproctitle("Sux Services " PACKAGE_VERSION SETPROCTITLE_APPEND);

    load_modules();

    log_set_irc_wrapper();
    start_master_thread();

    return EXIT_FAILURE;
}

static gint spawn(gint argc, gchar **argv)
{
    gint child_pid;
    GError *err;
    gchar **child_argv;
    gint i;

    log_set_tty_wrapper();

    child_argv = g_new0(gchar *, argc + 1);
    for(i = 0; i < argc; i++)
    {
	child_argv[i] = argv[i];
    }
    child_argv[i++] = "--nodetach";
    child_argv[i] = NULL;
	
    if(g_spawn_async_with_pipes(NULL, child_argv, NULL,  G_SPAWN_LEAVE_DESCRIPTORS_OPEN |
		G_SPAWN_STDOUT_TO_DEV_NULL | G_SPAWN_STDERR_TO_DEV_NULL | G_SPAWN_DO_NOT_REAP_CHILD,
		NULL, NULL, &child_pid, NULL, NULL, NULL, &err))
    {
	g_message("Services started in background (master thread id: %d)", child_pid);

	return EXIT_SUCCESS;
    }

    g_critical("Failed to start master thread: %s", err->message);
    g_error_free(err);

    return EXIT_FAILURE ;
}

gint main(gint argc, gchar **argv)
{
    gint (*func)(gint, gchar **);

    if(!GLIB_CHECK_VERSION(2, 2, 0))
    {
	g_critical("GLib version 2.2.0 or above is required");
	exit(EXIT_FAILURE);
    }

    me.boot = 0;
    parse_argc_argv(argc, argv);

    func = (options.nodetach) ? run : spawn;
    return (*func)(argc, argv);
}
