/*
 * Copyright 2002 Barnaba Marcello <vjt@azzurra.org>
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2a. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in
 *     the documentation and/or other materials provided with the
 *     distribution.
 * 2b. Redistribution in binary form requires specific prior written
 *     authorization of the maintainer.
 * 
 * 3. All advertising materials mentioning features or use of this
 *    software must display the following acknowledgement:
 *    This product includes software developed by Barnaba Marcello.
 * 
 * 4. The names of the maintainer, developers and contributors may not be
 *    used to endorse or promote products derived from this software
 *    without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE MAINTAINER, DEVELOPERS AND
 * CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL
 * THE DEVELOPERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 * ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

/* $Id: operserv.c,v 1.30 2003/03/26 21:18:00 vjt Exp $ */

#include "sux.h"
#include "operserv.h"
#include "help.h"
#include "sql.h"
#include "main.h"
#include "match.h"
#include "network.h"
#include "sux-printf.h"
#include "conf.h"
#include "modules.h"
#include "threads.h"

/* variables
 */
User *os, *noticer;
GMutex *os_akill_mutex;
GSList *active_akills = NULL;

REMOTE_TABLE_INSTANCE(user);
REMOTE_TABLE_INSTANCE(channel);

static MYSQL_RES *os_res;
static MYSQL_ROW os_row;

const gchar sux_module_name[] = "OperServ";
const gchar sux_module_version[] = "0.1.7";
const gchar sux_module_info[] = "Operator Services";

/* data types
 */
struct os_find_req
{
    User *requester;
    guint count;
    GPatternSpec *glob;
};

struct os_global_req
{
    gpointer data;
    gchar *msg;
};

/* private functions
 */
static void os_find_user(gchar *, User *, struct os_find_req *);
static void os_global_host(gchar *, User *, struct os_global_req *);
static void os_global_server(gchar *, User *, struct os_global_req *);
static void os_global_all(gchar *, User *, struct os_global_req *);
static gint os_print_tree(User *, GNode *, gint);

static void os_check_akill_ghfunc(gchar *, User *);
static gboolean os_check_akill(User *);
static void os_initial_synch(void);

/* for operserv entry point, look into
 * operserv-cmd.gperf
 */
gchar *g_module_check_init(GModule *mod)
{
    extern GSList *agents;
    Agent *os_a, *noticer_a;

    os_a = g_new0(Agent, 1);
    os_a->name = "OperServ";
    os_a->id = SYMBOL_SECTION_OPERSERV;
    os_a->u = &os;
    agents = g_slist_prepend(agents, (gpointer) os_a);

    noticer_a = g_new0(Agent, 1);
    noticer_a->name = "Global Noticer";
    noticer_a->id = SYMBOL_SECTION_NOTICER;
    noticer_a->u = &noticer;
    agents = g_slist_prepend(agents, (gpointer) noticer_a);

    agent_setcallback("OPERSERV", operserv);

    message_add_callback(MESSAGE_NEWNICK, SUX_MSG_PRIORITY_HIGH, (message_func) os_check_akill);
    message_add_callback(MESSAGE_OSREQ, SUX_MSG_PRIORITY_NORMAL, (message_func) operserv);
    
    if(me.boot)
    {
	agent_boot(os_a);
	agent_boot(noticer_a);

	os_initial_synch();
    }

    os_akill_mutex = g_mutex_new();
    os_akill_db_synch();

    g_message_syslog("MDL: Loaded OperServ module");

    return NULL;
}

void g_module_unload(GModule *mod)
{
    extern GSList *agents;
    GSList *agent;

    agent = g_slist_find_custom(agents, GINT_TO_POINTER(SYMBOL_SECTION_OPERSERV),
	    (GCompareFunc) agent_find);
    agents = g_slist_remove_link(agents, agent);
    agent_shutdown((Agent *)agent->data);
    g_free(agent->data);
    g_slist_free_1(agent);

    agent = g_slist_find_custom(agents, GINT_TO_POINTER(SYMBOL_SECTION_NOTICER),
	    (GCompareFunc) agent_find);
    agents = g_slist_remove_link(agents, agent);
    agent_shutdown((Agent *)agent->data);
    g_free(agent->data);
    g_slist_free_1(agent);

    g_mutex_lock(os_akill_mutex);
    while(active_akills)
    {
	AKill *a = active_akills->data;
	
	g_source_del(a->timeout);
	g_free(a->user);
	g_free(a->host);
	g_free(a);

	active_akills = g_slist_delete_link(active_akills, active_akills);
    }
    g_mutex_unlock(os_akill_mutex);
    g_mutex_free(os_akill_mutex);
    
    message_del_callback(MESSAGE_NEWNICK, (message_func) os_check_akill);
    message_del_callback(MESSAGE_OSREQ, (message_func) operserv);

    g_message_syslog("MDL: Unloaded OperServ module");
}

/* help wrapper around sux_help()
 */
void os_help(User *caller, gint parc, gchar **parv)
{
    sux_help(caller, os, OPERSERV_HELP_ID, parv ? parv[0] : NULL);
}

void os_find(User *caller, gint parc, gchar **parv)
{
    struct os_find_req rq;
    gchar mask[NICKLEN + 1 + USERLEN + 1 + HOSTLEN + 1];

    if(parc < 1)
    {
	send_out(":%s NOTICE %s :Not enough parameters",
		os->name, caller->name);
	return;
    }

    fixup_mask(parv[0], mask, sizeof(mask));

    rq.requester = caller;
    rq.count = 0;
    rq.glob = g_pattern_spec_new(mask);

    send_out(":%s NOTICE %s :Users matching %s",
	    os->name, caller->name, mask);
    
    g_hash_table_foreach(_TBL(user).__hash_tbl, (GHFunc) os_find_user, &rq);

    send_out(":%s NOTICE %s :End of search. Users found: %d",
	    os->name, caller->name, rq.count);

    g_pattern_spec_free(rq.glob);
}

void os_stats(User *caller, gint parc, gchar **parv)
{
    gint reg_chan, reg_nick;
    time_t up = NOW - me.boot;

    sql_query_assert(&os_res, "SELECT COUNT(`id`) FROM `nicks`");

    os_row = mysql_fetch_row(os_res);
    reg_nick = atoi(os_row[0]);
    mysql_free_result(os_res);

    sql_query_assert(&os_res, "SELECT COUNT(`id`) FROM `chans`");

    os_row = mysql_fetch_row(os_res);
    reg_chan = atoi(os_row[0]);
    mysql_free_result(os_res);

    send_out(":%s NOTICE %s :*** Services Stats ***",
	    os->name, caller->name);
    send_out(":%s NOTICE %s :Current users: %d, channels: %d",
	    os->name, caller->name, _TBL(user).count(), _TBL(channel).count());
    send_out(":%s NOTICE %s :Registered nicks: %d, chans: %d",
	    os->name, caller->name, reg_nick, reg_chan);
    send_out(":%s NOTICE %s :Services uptime: %ld days, %02ld:%02ld:%02ld",
	    os->name, caller->name, up / 86400, (up / 3600) % 24, (up / 60) % 60, up % 60);
    send_out(":%s NOTICE %s :*** End of Stats ***",
	    os->name, caller->name);
}

void os_uptime(User *caller, gint parc, gchar **parv)
{
    time_t up = NOW - me.boot;

    send_out(":%s NOTICE %s :Services uptime: %ld days, %02ld:%02ld:%02ld",
	    os->name, caller->name, up / 86400, (up / 3600) % 24, (up / 60) % 60, up % 60);
}

void os_global(User *caller, gint parc, gchar **parv)
{
    GHFunc out_func;
    gchar *target;
    struct os_global_req rq;
    gboolean free_pattern = FALSE;

    if(parc < 2)
    {
	send_out(":%s NOTICE %s :Not enough parameters",
		os->name, caller->name);
	return;
    }

    target = parv[0];
    rq.msg = parv[1];

    if(mycmp(target, "ALL") == 0)
    {
	out_func = (GHFunc) os_global_all;
	rq.data = NULL;
	target = "ALL";
    }
    else if((rq.data = (gpointer) _TBL(user).get(target)))
    {
	out_func = (GHFunc) os_global_server;
    }
    else
    {
	out_func = (GHFunc) os_global_host;
	rq.data = (gpointer) g_pattern_spec_new(target);
	free_pattern = TRUE;
    }

    send_out(":%s GLOBOPS :%s (%s@%s) used the global notice command ( target: %s )",
	    os->name, caller->name, caller->username, caller->host, target);

    g_hash_table_foreach(_TBL(user).__hash_tbl, out_func, (gpointer) &rq);

    if(free_pattern)
    {
	g_pattern_spec_free((GPatternSpec *)rq.data);
    }
}

void os_version(User *caller, gint parc, gchar **parv)
{
    struct utsname un;
    
    send_out(":%s NOTICE %s :*** Version info ***",
	    os->name, caller->name);
    send_out(":%s NOTICE %s :Sux Services - Core services",
	    os->name, caller->name);
    send_out(":%s NOTICE %s :Version: " PACKAGE_STRING,
	    os->name, caller->name);
    send_out(":%s NOTICE %s :Birth date: " __DATE__ " at " __TIME__,
	    os->name, caller->name);
    send_out(":%s NOTICE %s :mySQL Server version: %s",
	    os->name, caller->name, sql_get_server_info());

    if(uname(&un) < 0)
    {
	g_warning("Cannot get system info: uname(): %s",
		g_strerror(errno));
    }
    else
    {
	send_out(":%s NOTICE %s :Running on %s %s; host: %s; arch: %s",
		os->name, caller->name, un.sysname,
		un.release, un.nodename, un.machine);
    }

    send_out(":%s NOTICE %s :Coded and maintained by: " PACKAGE_BUGREPORT "",
	    os->name, caller->name);
    send_out(":%s NOTICE %s :*** End of version ***",
	    os->name, caller->name);
}

void os_map(User *caller, gint parc, gchar **parv)
{
    extern GNode *server_tree;
    gint count;

    send_out(":%s NOTICE %s :*** Network map ***",
	    os->name, caller->name);
    count = os_print_tree(caller, server_tree, 0);
    send_out(":%s NOTICE %s :-- Total servers: %d",
	    os->name, caller->name, count);
    send_out(":%s NOTICE %s :*** End of map ***",
	    os->name, caller->name);
}

/* private functions
 */
static gint os_print_tree(User *caller, GNode *subtree, gint depth)
{
    GNode *johnny; /* walker */
    gint count = 0;
    gint i;
    gchar *pipes, *s;
    
    johnny = g_node_get_root(subtree);
    if(depth)
    {
	pipes = g_new0(gchar, depth * 3 + 1);
	s = pipes;

	for(i = 0; i < depth; i++)
	{
	    *s++ = ' ';
	    *s++ = ' ';
	    *s++ = (johnny->parent && johnny->parent->children) ? '|' : ' ';

	    johnny = johnny->children;
	}
	*s = '\0';
    }
    else
    {
	pipes = "";
	s = NULL;
    }

    johnny = subtree;
    while(johnny)
    {
	if(s && !johnny->next)
	    *(s - 1) = '`';

	send_out(":%s NOTICE %s :%s--%c %s [%s]", os->name, caller->name,
		pipes, (johnny->children) ? '+' : '-',
		((User *)johnny->data)->name, ((User *)johnny->data)->info);

	count++;

	if(johnny->children)
	{
	    count += os_print_tree(caller, johnny->children, depth + 1);
	}

	johnny = johnny->next;
    }

    if(depth)
    {
	g_free(pipes);
    }

    return count;
}

static void os_find_user(gchar *dummy, User *u, struct os_find_req *rq)
{
    gchar match_vs[NICKLEN + 1 + USERLEN + 1 + HOSTLEN + 1];

    if(u->flags & (FLAGS_MYAGENT|FLAGS_SERVER|FLAGS_MYCLIENT))
    {
	return;
    }

    sux_snprintf(match_vs, sizeof(match_vs), "%s!%s@%s", u->name, u->username, u->host);

    if(g_pattern_match_string(rq->glob, match_vs))
    {
	send_out(":%s NOTICE %s :%d) %s (%s@%s) [%s]",
		os->name, rq->requester->name, ++rq->count, u->name,
		u->username, u->host, u->server->name);
    }

    return;
}

static void os_global_all(gchar *dummy, User *u, struct os_global_req *rq)
{
    if(u->flags & (FLAGS_MYAGENT|FLAGS_SERVER|FLAGS_MYCLIENT))
    {
	return;
    }

    send_out(":%s NOTICE %s :Global notice >> %s",
	    noticer->name, u->name, rq->msg);
}

static void os_global_server(gchar *dummy, User *u, struct os_global_req *rq)
{
    if(u->flags & (FLAGS_MYAGENT|FLAGS_SERVER|FLAGS_MYCLIENT))
    {
	return;
    }

    if(u->server == (User *) rq->data)
    {
	send_out(":%s NOTICE %s :Global notice >> %s",
		noticer->name, u->name, rq->msg);
    }
}

static void os_global_host(gchar *dummy, User *u, struct os_global_req *rq)
{
    if(u->flags & (FLAGS_MYAGENT|FLAGS_SERVER|FLAGS_MYCLIENT))
    {
	return;
    }

    if(g_pattern_match_string((GPatternSpec *)rq->data, u->host))
    {
	send_out(":%s NOTICE %s :Global notice >> %s",
		noticer->name, u->name, rq->msg);
    }
}

static void os_initial_synch(void)
{
    time_t synch_time;

    sql_lock_tables();

    synch_time = time(NULL);
    send_out(":%s GLOBOPS :Starting OperServ synch with database", me.name);
    g_hash_table_foreach(_TBL(user).__hash_tbl, (GHFunc) os_check_akill_ghfunc, NULL);
    send_out(":%s GLOBOPS :OperServ synch completed in %ld second(s)", me.name, time(NULL) - synch_time);

    sql_unlock_tables();
}

static void os_check_akill_ghfunc(gchar *dummy, User *u)
{
    os_check_akill(u);
}

static gboolean os_check_akill(User *u)
{
    gchar userathost[USERLEN + 1 + HOSTLEN + 1];
    gchar *user, *host;

    sux_snprintf(userathost, sizeof(userathost), "%s@%s", u->username, u->host);

    sql_query_assert_ret(FALSE, &os_res,
	    "SELECT `id`, `display_mask`, `reason`, `set_by`, "
	    "UNIX_TIMESTAMP(`time_set`), UNIX_TIMESTAMP(`expire`) "
	    "FROM `akill` WHERE '%s' LIKE `mask`",
	    userathost);

    if(!mysql_num_rows(os_res))
    {
	mysql_free_result(os_res);

	return TRUE;
    }

    os_row = mysql_fetch_row(os_res);
    /* +----+------+--------+--------+----------+--------+
     * | id | mask | reason | set_by | time_set | expire |
     * +----+------+--------+--------+----------+--------+
     */
    user = os_row[1];
    host = strchr(user, '@');
    *host++ = '\0';

    send_out("AKILL %s %s %ld %s %s :%s [AKill ID: %s]",
	    host, user, atoi(os_row[5]) - NOW, os_row[3], os_row[4], os_row[2], os_row[0]);

    /* stop the message
     */
    return FALSE;
}
