/*
 * Copyright 2002 Barnaba Marcello <vjt@azzurra.org>
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2a. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in
 *     the documentation and/or other materials provided with the
 *     distribution.
 * 2b. Redistribution in binary form requires specific prior written
 *     authorization of the maintainer.
 * 
 * 3. All advertising materials mentioning features or use of this
 *    software must display the following acknowledgement:
 *    This product includes software developed by Barnaba Marcello.
 * 
 * 4. The names of the maintainer, developers and contributors may not be
 *    used to endorse or promote products derived from this software
 *    without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE MAINTAINER, DEVELOPERS AND
 * CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL
 * THE DEVELOPERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 * ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

/* $Id: threads.c,v 1.21 2003/03/26 21:18:02 vjt Exp $ */

#include "sux.h"
#include "main.h"
#include "network.h"
#include "log.h"
#include "threads.h"
#include "parse.h"
#include "help.h"

#define ALLOCATED_NFDS	4

extern GMutex *net_readbuf_mutex;
extern GMutex *net_writebuf_mutex;
extern GMutex *net_tag_mutex;

extern GCond *net_readbuf_cond;

/* threads entry points
 */
static void network_thread(void);
static void parser_thread(void);
static void signal_thread(void);
static void master_thread(void);

/* setup functions
 */
static void setup_mutexes(void);
static void setup_signals(void);
static void unblock_fatal_sigs(void);

/* functions we call when we die
 */
static gint wait_for_termination(void);
static void join_threads(gint);

/* io-loop functions
 */
static G_INLINE_FUNC void my_g_main_context_iteration(void);

/* misc functions
 */
static void fatal_termination(gint);

static GThread *network_thread_ptr, *parser_thread_ptr, *signal_thread_ptr;
static GMainContext *ctx;
static GAsyncQueue *sig_queue = NULL, *thread_queue;

GLOBAL_RUN_DECLARE();

void start_master_thread(void)
{
    /* setup the mutexes for correct
     * multi-thread operation
     */
    setup_mutexes();

    /* setup the sendQ and recvQ
     */
    setup_netbuf();

    /* setup the user/channel tables, and 
     * the memory pools
     */
    setup_tables();

    /* load help data
     */
    setup_help_tables();

    master_thread();
}

static void join_threads(gint sig)
{
    STOP_RUNNING();
    
    g_cond_signal(net_readbuf_cond);

    raise(sig);
    if(sig != SIGSEGV && sig != SIGABRT)
    {
	g_thread_join(network_thread_ptr);
	g_thread_join(parser_thread_ptr);
    }

    /*
     * here we have only this thread running
     */
    while(net_active_tags())
    {
	my_g_main_context_iteration();
    }

    net_shutdown();

    sql_shutdown();
}

static void setup_mutexes(void)
{
    net_readbuf_mutex = g_mutex_new();
    net_readbuf_cond = g_cond_new();
    net_writebuf_mutex = g_mutex_new();
    net_tag_mutex = g_mutex_new();

    sig_queue = g_async_queue_new();

    GLOBAL_RUN_INIT();
}

static void fatal_termination(gint sig)
{
    g_message("Aieeeee!!! Ship sinks !!! Women and childrens first !!!");

    push_signal(&sig);

    abort();
}

static void setup_signals(void)
{
    sigset_t signal_set;
    
    sigfillset(&signal_set);

    pthread_sigmask(SIG_BLOCK, &signal_set, NULL);

    signal(SIGSEGV, fatal_termination);
}

static void unblock_fatal_sigs(void)
{
    sigset_t signal_set;

    sigaddset(&signal_set, SIGSEGV);
    sigaddset(&signal_set, SIGABRT);

    pthread_sigmask(SIG_UNBLOCK, &signal_set, NULL);
}

static gint wait_for_termination(void)
{
    gint *received_signal;
    
    received_signal = g_async_queue_pop(sig_queue);
    if(*received_signal != 0)
    {
	g_critical("Received signal %d (%s), quitting",
		*received_signal, g_strsignal(*received_signal));
    }

    return(*received_signal);
}

static G_INLINE_FUNC void my_g_main_context_iteration(void)
{
    gint max_priority;
    gint timeout;
    gint nfds;
    static GPollFD fds[ALLOCATED_NFDS];

    g_main_context_prepare(ctx, &max_priority);
    nfds = g_main_context_query(ctx, max_priority, &timeout, fds, ALLOCATED_NFDS);

    if(nfds || timeout != 0)
    {
	if(poll((struct pollfd *)fds, nfds, timeout) < 0 && errno != EINTR)
	{
	    g_errno_critical("poll()");
	}
    }

    g_main_context_check(ctx, max_priority, fds, nfds);
    g_main_context_dispatch(ctx);

    return;
}

static void master_thread(void)
{
    GError *err = NULL;
    gpointer thr_id;
    gint exit_sig;

    setup_signals();
    thread_queue = g_async_queue_new();

    signal_thread_ptr = g_thread_create_full((GThreadFunc)signal_thread,
	    NULL, 0, FALSE, TRUE, G_THREAD_PRIORITY_NORMAL, &err);
    if(signal_thread_ptr == NULL)
    {
	exit(EXIT_FAILURE);
    }
    thr_id = g_async_queue_pop(thread_queue);
    g_message_syslog("THR: Initialized signal thread, pid: %d", GPOINTER_TO_INT(thr_id));

    network_thread_ptr = g_thread_create_full((GThreadFunc)network_thread,
	    NULL, 0, TRUE, TRUE, G_THREAD_PRIORITY_NORMAL, &err);
    if(network_thread_ptr == NULL)
    {
	exit(EXIT_FAILURE);
    }
    thr_id = g_async_queue_pop(thread_queue);
    g_message_syslog("THR: Initialized network thread, pid: %d", GPOINTER_TO_INT(thr_id));

    parser_thread_ptr = g_thread_create_full((GThreadFunc)parser_thread,
	    NULL, 0, TRUE, TRUE, G_THREAD_PRIORITY_NORMAL, &err);
    if(parser_thread_ptr == NULL)
    {
	exit(EXIT_FAILURE);
    }
    thr_id = g_async_queue_pop(thread_queue);
    g_message_syslog("THR: Initialized parser/dispatcher thread, pid: %d", GPOINTER_TO_INT(thr_id));

    g_async_queue_unref(thread_queue);

    exit_sig = wait_for_termination();
    join_threads(exit_sig);
}

static void network_thread(void)
{
    THREAD_RUN_DECLARE();

    unblock_fatal_sigs();
    sql_thread_init();

    ctx = g_main_context_new();
    g_main_context_ref(ctx);

    time(&me.boot);

    net_add_tags();
    irc_nego_start();

    g_async_queue_push(thread_queue, GINT_TO_POINTER(getpid()));
 
    THREAD_RUN_CHECK();

    while(THREAD_IS_RUNNING())
    {
	my_g_main_context_iteration();

	THREAD_RUN_CHECK();
    }

    g_thread_exit(NULL);
}

static void parser_thread(void)
{
    extern GString *recvQ;
    const gint STRINGS_PER_CYCLE = 2048;
    gchar **strings = g_new0(gchar *, STRINGS_PER_CYCLE);
    GString *read_data = g_string_sized_new(READBUFSZ);
    gint i, count;
    THREAD_RUN_DECLARE();

    unblock_fatal_sigs();
    sql_thread_init();

    g_async_queue_push(thread_queue, GINT_TO_POINTER(getpid()));

    THREAD_RUN_CHECK();

    while(THREAD_IS_RUNNING())
    {
	g_mutex_lock(net_readbuf_mutex);
	if(!recvQ->len)
	{
	    g_cond_wait(net_readbuf_cond, net_readbuf_mutex);
	}

	if(!recvQ->len)
	{
	    /*
	     * possibly are we terminated ?
	     */
	    g_mutex_unlock(net_readbuf_mutex);
	    
	    THREAD_RUN_CHECK();
	    continue;
	}

	read_data = g_string_assign(read_data, recvQ->str);
	if(my_g_strsplit(read_data->str, '\n', STRINGS_PER_CYCLE, &count, strings) == FALSE)
	{
	    /* there is something to stick back into the recvQ */
	    recvQ = g_string_assign(recvQ, strings[--count]);
	}
	else
	{
	    recvQ = g_string_erase(recvQ, 0, -1);
	}
	g_mutex_unlock(net_readbuf_mutex);

	NOW = time(NULL);

	for(i = 0; i < count; i++)
	{
	    parse(strings[i]);
	}

	THREAD_RUN_CHECK();
    }

    sql_cleanup_before_exit();
    g_thread_exit(NULL);

}

static void signal_thread(void)
{
    sigset_t signal_set;
    gint *sig = g_new0(gint, 1);
    
    sigfillset(&signal_set);
    sigdelset(&signal_set, SIGSEGV);
    sigdelset(&signal_set, SIGABRT);

    g_async_queue_push(thread_queue, GINT_TO_POINTER(getpid()));

    for(;;)
    {
	sux_sigwait(&signal_set, sig);

	switch(*sig)
	{
	    case SIGTERM:
	    case SIGQUIT:
	    case SIGHUP:
	    case SIGINT:
		g_async_queue_push(sig_queue, sig);
		g_thread_exit(NULL);

		break;

		break;

	    case SIGPIPE:
	    default:
		break;
	}

	sigaddset(&signal_set, *sig);
    }

    g_thread_exit(NULL);
}

G_INLINE_FUNC GSource *g_source_add(GIOChannel *handle, GIOCondition cond, GIOFunc callback)
{
    GSource *gs;

    gs = g_io_create_watch(handle, cond);
    g_source_set_priority(gs, G_PRIORITY_DEFAULT);
    g_source_set_callback(gs, (GSourceFunc) callback, NULL, NULL);

    g_source_attach(gs, ctx);

    return gs;
}

G_INLINE_FUNC GSource *g_timeout_source_add(guint interval, GSourceFunc callback, gpointer user_data)
{
    GSource *gs;

    gs = g_timeout_source_new(interval);
    g_source_set_priority(gs, G_PRIORITY_DEFAULT);
    g_source_set_callback(gs, (GSourceFunc) callback, user_data, NULL);

    g_source_attach(gs, ctx);

    return gs;
}

G_INLINE_FUNC void g_source_del(GSource *gs)
{
    g_return_if_fail(gs != NULL);
    
    g_source_destroy(gs);

    g_source_unref(gs);
}

void push_signal(gint *signum)
{
    if(sig_queue)
    {
	g_async_queue_push(sig_queue, signum);
    }
}
